﻿# phy_tja1100.py
#
# This file is part of python wrapper for the BROADWAY2 package.
#
# MIT License

# Copyright (c) 2013-2018 FibreCode GmbH <info@fibrecode.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import array
from broadway2 import broadway_api, phy_access

from adi import ADIN1100Register

class PhyADIN1100:

    MMD1_DEV_ID1 = 0x0283
    MMD1_DEV_ID2_OUI = 0x2F

    SpeedSelect = {4: '10BASE-T1L', 1: '100 MBit/s', 2: '1000 MBit/s', 3: 'reserved', 7: 'n/a'}

    def __init__(self, phyA):
        """Default constructor"""
        self.__phyAccess = phyA
        self.Register = ADIN1100Register()

    def get_register(self):
        return ADIN1100Register.Register

    def get_register_sorted(self):
        res = []
        for reg in ADIN1100Register.Register:
            res.append((reg, ADIN1100Register.Register[reg]))
            
        return sorted(res)

    def get_reg_devs(self):
        devs = {}
        for devreg in ADIN1100Register.Register:
            dev = devreg >> 16
            try:
                tst = devs[dev]
            except:
                devs[dev] = 1
                
        return sorted(list(devs))



    def get_register_details(self, regIndex):
        try:
            regBits = ADIN1100Register.RegisterBits[regIndex]
        except:
            regBits = None
             
        return regBits

    def lookup_register(self, regIndex):
        """lookup register offset and return tag"""
        try:
            tag = ADIN1100Register.Register[regIndex]['Tag']
        except:
            tag = 'N/A'
        return tag

    def lookup_reg_symbol(self, regIndex, symbol):
        """dynamic lookup to save huge dic"""
        regBits = ADIN1100Register.RegisterBits[regIndex]
        idx = 0
        for offset in regBits:
            if symbol == regBits[offset]['Symbol']:
                return (idx, offset, regBits[offset]['Width'])
            idx += 1
        return None
         
    def get_device(self):
        """helper function to return current used device"""
        return self.__phyAccess
        
    def connect(self, phyA):
        """connect to given device instance"""
        self.__phyAccess = phyA

    def disconnect(self):
        """disconnect this instance from current USB HID device"""
        self.__phyAccess = None

    @property
    def isConnected(self):
        """return True/False if phy instance is connected to USB-Adapter"""
        return self.__phyAccess != None

    def decode(self, reg, value):
        """generic bit decoder, based on register/Bit-Definitions"""
        res = []
        try:
            regDef = ADIN1100Register.RegisterBits[reg]
        except:
            return res

        for bits in regDef:
            try:
                bitDef = regDef[bits]
                mask = ((1 << (bits + bitDef['Width'])) - 1) & ~((1 << bits)-1)
            except:
                return res
            bitsValue = (value & mask) >> bits
            try:               
                if bitDef['Width'] == 1:
                    if bitsValue:
                        res.append((bits, bitDef['Width'], bitDef['Symbol'], bitDef['Values'][1], bitsValue))
                    else:
                        res.append((bits, bitDef['Width'],bitDef['Symbol'], bitDef['Values'][0], bitsValue))
                else:
                    if len(bitDef['Values']) > 0:
                        res.append((bits, bitDef['Width'], bitDef['Symbol'], bitDef['Values'][bitsValue], bitsValue))  # enum
                    else:
                        res.append((bits, bitDef['Width'], bitDef['Symbol'], '', bitsValue))   # just use integer value
            except:
                res.append((bits, 0, '', '', -1))
        return sorted(res, reverse=True)

    def decode_symbol(self, reg, symbol, value):
        """bit decoder for only one known symbol"""
        try:
            regDef = ADIN1100Register.RegisterBits[reg]
        except:
            return None

        for bits in regDef:
            if regDef[bits]['Symbol'] == symbol:
                try:
                    bitDef = regDef[bits]
                    mask = ((1 << (bits + bitDef['Width'])) - 1) & ~((1 << bits)-1)
                except:
                    return None
                bitsValue = (value & mask) >> bits
                try:               
                    if bitDef['Width'] == 1:
                        if bitsValue:
                            return (bits, bitDef['Width'], bitDef['Symbol'], bitDef['Values'][1], bitsValue)
                        else:
                            return (bits, bitDef['Width'],bitDef['Symbol'], bitDef['Values'][0], bitsValue)
                    else:
                        if len(bitDef['Values']) > 0:
                            return (bits, bitDef['Width'], bitDef['Symbol'], bitDef['Values'][bitsValue], bitsValue)  # enum
                        else:
                            return (bits, bitDef['Width'], bitDef['Symbol'], '', bitsValue)   # just use integer value
                except:
                    return (bits, 0, '', '', -1)

    def an_restart(self):
        regVal = self.read_register(ADIN1100Register.regAN_CONTROL)
        regVal |= (1<<9)
        self.write_register(TJA1100Register.regAN_CONTROL, regVal)


    @property
    def Reset(self):
        regVal = self.read_register(ADIN1100Register.regBCR)
        return (regVal & 0x8000) == 0x8000

    @Reset.setter
    def Reset(self, value):
        regVal = self.read_register(ADIN1100Register.regBCR)
        if value:
            regVal |= 0x8000
        else:
            regVal &= ~0x8000
        self.write_register(TJA1100Register.regBCR, regVal)

    @property
    def PCSLoopback(self):
        regVal = self.read_register(ADIN1100Register.regMAC_IF_LOOPBACK)
        return self.decode_symbol(ADIN1100Register.regMAC_IF_LOOPBACK, 'MAC_IF_LB_EN', regVal)[4] == 1

    @PCSLoopback.setter
    def PCSLoopback(self, value):
        regVal = self.read_register(ADIN1100Register.regMAC_IF_LOOPBACK)
        if value:
            regVal |= 1
        else:
            regVal &= ~1
        self.write_register(ADIN1100Register.regMAC_IF_LOOPBACK, regVal)


    @property
    def B10LLoopback(self):
        regVal = self.read_register(ADIN1100Register.regB10L_PMA_CNTRL)
        return self.decode_symbol(ADIN1100Register.regB10L_PMA_CNTRL, 'B10L_LB_PMA_LOC_EN', regVal)[4] == 1

    @B10LLoopback.setter
    def B10LLoopback(self, value):
        regVal = self.read_register(ADIN1100Register.regB10L_PMA_CNTRL)
        if value:
            regVal |= 1
        else:
            regVal &= ~1
        self.write_register(ADIN1100Register.regB10L_PMA_CNTRL, regVal)
    
    @property
    def MacLoopback(self):
        regVal = self.read_register(ADIN1100Register.regMAC_IF_LOOPBACK)
        return self.decode_symbol(ADIN1100Register.regMAC_IF_LOOPBACK, 'MAC_IF_LB_EN', regVal)[4] == 1

    @MacLoopback.setter
    def MacLoopback(self, value):
        regVal = self.read_register(ADIN1100Register.regMAC_IF_LOOPBACK)
        if value:
            regVal |= (1<<0)
        else:
            regVal &= ~(1<<0)
        self.write_register(ADIN1100Register.regMAC_IF_LOOPBACK, regVal)

    @property
    def MacRemoteLoopback(self):
        regVal = self.read_register(ADIN1100Register.regMAC_IF_LOOPBACK)
        return self.decode_symbol(ADIN1100Register.regMAC_IF_LOOPBACK, 'MAC_IF_REM_LB_EN', regVal)[4] == 1

    @MacRemoteLoopback.setter
    def MacRemoteLoopback(self, value):
        regVal = self.read_register(ADIN1100Register.regMAC_IF_LOOPBACK)
        if value:
            regVal |= (1<<2)
        else:
            regVal &= ~(1<<2)
        self.write_register(ADIN1100Register.regMAC_IF_LOOPBACK, regVal)
        

    @property
    def NetworkSpeed(self):
        regVal = self.read_register(ADIN1100Register.regPMA_PMD_BT1_ABILITY)
        speedCode = regVal & 0x7
        return self.SpeedSelect[speedCode]


    @property
    def AnComplete(self):
        regVal = self.read_register(ADIN1100Register.regAN_STATUS)
        return self.decode_symbol(ADIN1100Register.regAN_STATUS, 'AN_COMPLETE', regVal)[3]

    @property
    def AnLinkGood(self):
        regVal = self.read_register(ADIN1100Register.regAN_STATUS_EXTRA)
        return self.decode_symbol(ADIN1100Register.regAN_STATUS_EXTRA, 'AN_LINK_GOOD', regVal)[3]

    @property
    def AnLinkStatus(self):
        regVal = self.read_register(ADIN1100Register.regAN_STATUS)
        return self.decode_symbol(ADIN1100Register.regAN_STATUS, 'AN_LINK_STATUS', regVal)[3]

    @property
    def AnMSConfigResolution(self):
        regVal = self.read_register(ADIN1100Register.regAN_STATUS_EXTRA)
        return self.decode_symbol(ADIN1100Register.regAN_STATUS_EXTRA, 'AN_MS_CONFIG_RSLTN', regVal)[3]

    @property
    def EnableV24(self):
        regVal = self.read_register(ADIN1100Register.regAN_ADV_ABILITY_H)
        if (regVal & (1<<12)) and (regVal & (1<<13)):
            return True
        return False

    @EnableV24.setter
    def EnableV24(self, value):
        regVal = self.read_register(ADIN1100Register.regAN_ADV_ABILITY_H)
        if value:
            regVal |= (3<<12)
        else:
            regVal &= ~(3<<12)
        self.write_register(ADIN1100Register.regAN_ADV_ABILITY_H, regVal)


    @property
    def b10TxOutputDisable(self):
        regVal = self.read_register(ADIN1100Register.regB10L_PMA_CNTRL)
        return self.decode_symbol(ADIN1100Register.regB10L_PMA_CNTRL, 'B10L_TX_DIS_MODE_EN', regVal)[4] == 1

    @b10TxOutputDisable.setter
    def b10TxOutputDisable(self, value):
        regVal = self.read_register(ADIN1100Register.regB10L_PMA_CNTRL)
        if value:
            regVal |= (1 << 14)
        else:
            regVal &= ~(1 << 14)
        self.write_register(ADIN1100Register.regB10L_PMA_CNTRL, regVal)

    # AnForceMaster
    @property
    def AnAdvForceMaster(self):
        regVal = self.read_register(ADIN1100Register.regAN_ADV_ABILITY_L)
        return self.decode_symbol(ADIN1100Register.regAN_ADV_ABILITY_L, 'AN_ADV_FORCE_MS', regVal)[4] & 1

    @AnAdvForceMaster.setter
    def AnAdvForceMaster(self, value):
        regVal = self.read_register(ADIN1100Register.regAN_ADV_ABILITY_L)
        if value:
            regVal |= (1<<12)
        else:
            regVal &= ~(1<<12)
        self.write_register(ADIN1100Register.regAN_ADV_ABILITY_L, regVal)

    # AnAdvMaster
    @property
    def AnAdvMaster(self):
        regVal = self.read_register(ADIN1100Register.regAN_ADV_ABILITY_M)
        return self.decode_symbol(ADIN1100Register.regAN_ADV_ABILITY_M, 'AN_ADV_MST', regVal)[4] & 1

    @AnAdvMaster.setter
    def AnAdvMaster(self, value):
        regVal = self.read_register(ADIN1100Register.regAN_ADV_ABILITY_M)
        if value:
            regVal |= (1<<4)
        else:
            regVal &= ~(1<<4)
        self.write_register(ADIN1100Register.regAN_ADV_ABILITY_M, regVal)


    @property
    def B10LLinkStatus(self):
        regVal = self.read_register(ADIN1100Register.regB10L_PMA_LINK_STAT)
        return self.decode_symbol(ADIN1100Register.regB10L_PMA_LINK_STAT, 'B10L_LINK_STAT_OK', regVal)[3]

    @property
    def AutoMode(self):
        return True

    @AutoMode.setter
    def AutoMode(self, value):
        pass

    def get_b10l_loopback_mode(self):
        if self.PCSLoopback:
            return 1
        elif self.B10LLoopback:
            return 2
        elif self.MacRemoteLoopback:
            return 3

        return 0

    def enable_loopback_mode(self, loopbackMode):

        self.PCSLoopback = False
        self.MacLoopback = False
        self.MacRemoteLoopback = False
       
        if loopbackMode == 1:
            self.PCSLoopback = True
        elif loopbackMode == 2:
            self.MacLoopback = True
        elif loopbackMode == 3:
            self.MacRemoteLoopback = True

    def disable_loopback_mode(self):
        self.PCSLoopback = False
        self.MacLoopback = False
        self.MacRemoteLoopback = False


    def get_b10l_test_mode(self):
        """read current testMode, no action/change"""

        # read test-mode
        regVal = self.read_register(ADIN1100Register.regB10L_TEST_MODE_CNTRL)

        testMode = (regVal >> 13) & 0x7

        return testMode
        
        
    def enable_b10l_test_mode(self, testMode):
        """bring Phy to one of selected Test-Modes 1-3
        return:
            True: if test-mode successfully activated
            False: Wrong arg or test-mode activation failed
        """

        if testMode < 1 or testMode > 3:
            return False
        # disable power
        regVal = self.read_register(ADIN1100Register.regCRSM_SFT_PD_CNTRL)
        regVal |= 1
        self.write_register(ADIN1100Register.regCRSM_SFT_PD_CNTRL, regVal)
        # verify power-down
        regVal = self.read_register(ADIN1100Register.regCRSM_STAT)
        # verify PD_READY
        if (regVal & 0x0002) != 0x0002:
            return False
        # Disable AUTO-NEG
        regVal = self.read_register(ADIN1100Register.regAN_CONTROL)
        regVal &= ~(1<<12)  # clear AN_EN
        self.write_register(ADIN1100Register.regAN_CONTROL, regVal)

        # Enable AN Forced 
        regVal = self.read_register(ADIN1100Register.regAN_FRC_MODE_EN)
        regVal |= 1
        self.write_register(ADIN1100Register.regAN_FRC_MODE_EN, regVal)

        # config test-mode
        regVal = self.read_register(ADIN1100Register.regB10L_TEST_MODE_CNTRL)
        regVal &= ~(7<<13)  # clear
        regVal |= (testMode << 13)
        self.write_register(ADIN1100Register.regB10L_TEST_MODE_CNTRL, regVal)

        # finally enable power with testMode active
        regVal = self.read_register(ADIN1100Register.regCRSM_SFT_PD_CNTRL)
        regVal &= ~1
        self.write_register(ADIN1100Register.regCRSM_SFT_PD_CNTRL, regVal)

    def disable_b10l_test_mode(self):
        """bring Phy back to normal
        return:
            True: Phy back to normal, test-mode disabled
            False: Test-mode deactivation failed
        """

        # disable power
        regVal = self.read_register(ADIN1100Register.regCRSM_SFT_PD_CNTRL)
        regVal |= 1
        self.write_register(ADIN1100Register.regCRSM_SFT_PD_CNTRL, regVal)
        # verify power-down
        regVal = self.read_register(ADIN1100Register.regCRSM_STAT)
        # verify PD_READY
        if (regVal & 0x0002) != 0x0002:
            return False
        # Enable AUTO-NEG
        regVal = self.read_register(ADIN1100Register.regAN_CONTROL)
        regVal |= (1<<12)  # set AN_EN
        self.write_register(ADIN1100Register.regAN_CONTROL, regVal)

        # Disable AN Forced 
        regVal = self.read_register(ADIN1100Register.regAN_FRC_MODE_EN)
        regVal &= ~1
        self.write_register(ADIN1100Register.regAN_FRC_MODE_EN, regVal)

        # test-mode to 0
        regVal = self.read_register(ADIN1100Register.regB10L_TEST_MODE_CNTRL)
        regVal &= ~(7<<13)  # clear
        self.write_register(ADIN1100Register.regB10L_TEST_MODE_CNTRL, regVal)

        # finally enable power with testMode active
        regVal = self.read_register(ADIN1100Register.regCRSM_SFT_PD_CNTRL)
        regVal &= ~1
        self.write_register(ADIN1100Register.regCRSM_SFT_PD_CNTRL, regVal)

#
# private functions, take care to use them directly. Better use high-level API to access properties, not registers!
#
    def read_dev_register(self, dev, register):
        res = self.__phyAccess.read_dev_reg(dev, register)
        return res

    def write_dev_register(self, dev, register, regValue):
        res = self.__phyAccess.write_dev_reg(dev, register, regValue)
        return res

    def read_register(self, devreg):
        dev = (devreg >> 16) & 0xFFFF
        register = devreg & 0xFFFF
        res = self.__phyAccess.read_dev_reg(dev, register)
        return res

    def write_register(self, devreg, regValue):
        dev = (devreg >> 16) & 0xFFFF
        register = devreg & 0xFFFF
        res = self.__phyAccess.write_dev_reg(dev, register, regValue)
        return res
